import {ElementRef, Injectable, OnDestroy} from '@angular/core';
import {Observable, of, Subject, Subscription} from 'rxjs';
import {XplgContainerDiagramRestService} from './xplg-container-diagram-rest.service';
import {debounceTime, switchMap} from 'rxjs/operators';
import {CommunicationEvent, Theme} from '../models/communication-event';
import {XplgContainerDiagramEvents as DiagramEvents} from '../components/xplg-container-diagram/xplg-container-diagram';

@Injectable()
export class CommunicationService implements OnDestroy {
    data: any;
    _diagramType = '';
    restConfig$ = new Subject();

    private _theme: Theme = 'default';
    private _selectedNodeId: number = -1;
    private incomingMessage$ = new Subject<CommunicationEvent>();
    private targetWindow: ElementRef<HTMLIFrameElement> | null = null;
    private subscription: Subscription | null = null;

    constructor(private rest: XplgContainerDiagramRestService) {
        //TODO: migrate to onInit
        this.subscribeEmbeddedInstance();
    }

    set theme(theme: Theme) {
        this._theme = theme;
    }

    get theme(): Theme {
        return this._theme;
    }

    set selectedNodeId(id: number) {
        this._selectedNodeId = id;
    }

    get selectedNodeId(): number {
        return this._selectedNodeId;
    }

    set target(targetWindow: ElementRef<HTMLIFrameElement>) {
        this.targetWindow = targetWindow;
    }

    set diagramType(type: string) {
        this._diagramType = type;
    }

    get diagramType(): string {
        return this._diagramType;
    }

    embeddedInstanceAction(event: CommunicationEvent) {
        if (event) {
            this.incomingMessage$.next(event);
        }
    }

    fireEvent(event: CommunicationEvent) {
        if (event) {
            this.targetWindow?.nativeElement.contentWindow?.postMessage(
                {
                    ...event,
                    theme: this.theme,
                    selectedNodeId: this.selectedNodeId
                }, '*');
        }
    }

    private subscribeEmbeddedInstance() {
        this.subscription = this.incomingMessage$
            .pipe(
                debounceTime(100),
                switchMap(event => {
                      return this.eventsHandler(event);
                })
            )
            .subscribe(event => this.fireEvent(event));
    }

    private eventsHandler(event: CommunicationEvent): Observable<any> {
        if (event.callback) {
            event.callback();
        }
        switch (event.type) {
            case DiagramEvents.LOADED:
                return this.onLoad();
            case DiagramEvents.ADD_NODE:
                return this.addNode();
            case DiagramEvents.REMOVE_NODE:
                return this.removeNode();
            case DiagramEvents.REFRESH:
                // refresh embedded DIAGRAM
                return of(null);
            case DiagramEvents.INJECT_DATA:
                // inject embedded DIAGRAM data
                return of(null);
            case DiagramEvents.REQUEST_DATA:
                // request embedded DIAGRAM data
                return of(null);
            case DiagramEvents.DESTROY:
                // destroy embedded DIAGRAM
                return of(null);
            case DiagramEvents.UI_EVENT:
                // destroy embedded DIAGRAM
                return this.uiEventsHandler(event);
            default:
                return this.defaultAction();
        }
    }

    private uiEventsHandler(event: any): Observable<CommunicationEvent> {
        switch (event.params.type) {
            case 'click':
                const selectedNodeData = event.params?.data;
                this.restConfig$.next({rest: selectedNodeData?.rest, selectedNodeId: selectedNodeData.id});
                return of();
            case 'dragndrop':
                return of();
            default:
              return of();
        }
    }

    private onLoad(): Observable<CommunicationEvent> {
        return this.rest.embeddedDiagramAction(this.data, this.diagramType);
    }

    private addNode(): Observable<CommunicationEvent> {
        return this.rest.addNode();
    }

    private removeNode(): Observable<CommunicationEvent> {
        return this.rest.removeNode();
    }

    private defaultAction(): Observable<CommunicationEvent> {
        return of();
    }

    ngOnDestroy() {
        this.subscription?.unsubscribe();
    }
}
