import {
    AfterViewInit,
    Component, ElementRef,
    OnInit, Renderer2,
    ViewChild,
    ViewEncapsulation
} from '@angular/core';
import {
    DataSourceModel,
    Diagram,
    DiagramConstraints,
    IClickEventArgs,
    LayoutModel,
    NodeConstraints,
    NodeModel, PointPortModel,
    SelectorConstraints,
    SnapConstraints,
    SnapSettingsModel,
    TreeInfo
} from '@syncfusion/ej2-angular-diagrams';
import {DataManager} from '@syncfusion/ej2-data';
import {IEmbeddedSyncfusionComponent} from './iembedded-diagram-component';
import {NarrowLayoutService} from './data-transformers/narrow-layout.service';
import {AdvancedLayoutService} from './data-transformers/advanced-layout.service';
import {AdvancedLayoutAsyncService} from './data-transformers/advanced-layout-async.service';
import {MockDataProviderService} from './data-providers/mock-data-provider.service';
import {environment} from '../environments/environment';
import {NodeUtilsService} from './services/node-utils.service';
import {NodeHelper} from "./helpers/node-helper";
import {XPLGPaletteDefaultDark, XPLGPaletteDefaultLight} from "./constants/theme.constants";

export interface NodeInfo {
    id: string;
    value: string;
    role: string;
    amount: string;
    children: string[];
    offsetY: number;
    child: boolean;
    time: string;
    position: string;
    ports?: PointPortModel[];
    uiPrefs: {
        color: string;
        iconBackground: string;
        icon: string;
        tooltip: {
            description: string,
            events: string
        },
        hidden?: boolean
    },
    uiUtils: {
        color: string;
    }
}

export interface DataInfo {
    [key: string]: string;
}

export interface IXplgNode {
}

type LayoutType = 'narrow' | 'advanced';

@Component({
    selector: 'app-diagram-root',
    templateUrl: './syncfusion-impl.component.html',
    styleUrls: ['./syncfusion-impl.component.scss'],
    encapsulation: ViewEncapsulation.None,
    providers: [NodeUtilsService]
})
export class SyncfusionImplComponent implements OnInit, AfterViewInit, IEmbeddedSyncfusionComponent {
    diagramWidth = window.innerWidth;
    diagramHeight = window.innerHeight;
    constraints = DiagramConstraints.Default;
    selectedItems = {constraints: SelectorConstraints.None};
    maxLabelLength = new Map([
        [1, 76],
        [2, 40],
        [3, 24],
        [4, 16],
        [5, 12]
    ]);

    snapSettings: SnapSettingsModel = {
        constraints: SnapConstraints.None
    };

    selectedNodeId: number | undefined;
    selectedNode: HTMLElement | undefined;
    dataSourceSettings: DataSourceModel = {};
    loaded = true;
    ready = true;
    asyncDiagram = false;

    connectorDefaults = NarrowLayoutService.connectorDefaults;
    nodeDefaults = NarrowLayoutService.nodeDefault;

    @ViewChild('layout') layout: LayoutModel | undefined;
    @ViewChild('diagram') diagram: Diagram | undefined;

    constructor(private el: ElementRef,
                private renderer: Renderer2,
                private mock: MockDataProviderService,
                private nodeUtils: NodeUtilsService) {
    }

    ngOnInit(): void {
        if (environment.production) {
            window.parent.postMessage({type: 'loaded'}, '*');
            window.addEventListener('message', (message) => {
                if (message.data.type === 'injectData') {
                    this.applyConf(message.data.data);
                }
                if (message.data.type === 'toggleNode') {
                    if (!message.data.state && this.selectedNode) {
                        this.renderer.removeClass(this.selectedNode, 'active');
                    }
                    if (typeof message.data.targetId === 'string') {
                        this.selectNodeById(message.data.targetId);
                    }
                }
                if (message.data.theme) {
                    this.applyTheme(message.data.theme);
                }
                this.selectedNodeId = message.data.selectedNodeId;
            });
        } else {
            const mockDiagramData = this.mock.fetchDiagramData();
            const mockData = mockDiagramData.data;
            this.applyConf(mockData);
        }
    }

    ngAfterViewInit(): void {
        // @ts-ignore
        this.diagram.dataLoaded?.subscribe(() => {
            const delay = setTimeout(() => {
                this.prepareNodes();
                this.ready = true;
                clearTimeout(delay);
            });
        });
    }

    applyConf(diagramConf: any): void {
        this.ready = false;
        this.asyncDiagram = diagramConf?.async
        switch (diagramConf.diagramType) {
            case 'narrow':
                this.connectorDefaults = NarrowLayoutService.connectorDefaults;
                this.nodeDefaults = NarrowLayoutService.nodeDefault;
                break;
            case 'advanced':
                if (diagramConf.async) {
                    diagramConf = NodeHelper.addHiddenNodes(diagramConf);
                    this.connectorDefaults = AdvancedLayoutAsyncService.connectorDefaults;
                    this.nodeDefaults = AdvancedLayoutAsyncService.nodeDefault;
                } else {
                    this.connectorDefaults = AdvancedLayoutService.connectorDefaults;
                    this.nodeDefaults = AdvancedLayoutService.nodeDefault;
                }
        }
        this.bootstrapDiagram(diagramConf);
    }

    bootstrapDiagram(diagramConf: any): void {
        let nodesAmount = diagramConf.data.length;
        let marginTop = 20;
        const marginLeft = 20;
        const marginRight = 20;
        const maxLength = this.maxLabelLength.get(nodesAmount) || 10;
        let nodeWidth = maxLength * 10;
        let verticalSpacing = (this.el.nativeElement.offsetWidth - (nodeWidth * nodesAmount)) / (nodesAmount - 1);
        if (diagramConf?.async) {
            marginTop = 0;
            nodeWidth = maxLength * 10 + 36;
            nodesAmount = (nodesAmount + 1) / 2;
            verticalSpacing = ((this.el.nativeElement.offsetWidth - (nodeWidth * nodesAmount)) / (nodesAmount - 1) * 0.3);
        }
        if (verticalSpacing < 10) {
            verticalSpacing = 10;
        }
        this.diagramWidth = window.innerWidth;
        this.diagramHeight = window.innerHeight;
        this.layout = {
            type: 'HierarchicalTree',
            orientation: 'LeftToRight',
            horizontalAlignment: 'Left',
            verticalSpacing,
            verticalAlignment: 'Top',
            margin: {
                top: marginTop,
                left: marginLeft,
                right: marginRight
            },
            getLayoutInfo: (node: NodeModel, options: TreeInfo) => {
                options.type = 'Left';
                options.orientation = 'Horizontal';
            }
        };
        this.dataSourceSettings = {
            id: 'id',
            parentId: 'parent',
            dataManager: new DataManager(diagramConf.data),
            doBinding: (nodeModel: NodeModel, data: DataInfo, diagram: Diagram) => {
                let value = (nodeModel.data as NodeInfo).value;
                let amount = (nodeModel.data as NodeInfo).amount;
                // @ts-ignore
                const tooltip = data['uiPrefs']['tooltip'];
                if (amount?.length) {
                    (nodeModel.data as NodeInfo).amount = `(${amount})`;
                }
                if (value.length > maxLength) {
                    const defaultStartPosition = maxLength / 2;
                    (nodeModel.data as NodeInfo).amount = '';
                    (nodeModel.data as NodeInfo).value = `${value.slice(0, defaultStartPosition)}...${value.slice(value.length - defaultStartPosition, value.length)}`;
                }
                if (tooltip) {
                    const content = tooltip?.description;
                    const eventsAlias = parseInt(tooltip?.events) > 1 ? 'events' : 'event';
                    nodeModel.tooltip = {
                        content: `<span class="node-tooltip-content">${content}</span>${tooltip?.events} ${eventsAlias}`,
                        position: 'RightCenter',
                        relativeMode: 'Object'
                    };
                    nodeModel.constraints = (NodeConstraints.Default | NodeConstraints.Tooltip | NodeConstraints.ReadOnly)
                        & ~NodeConstraints.Drag & ~NodeConstraints.Select;
                }
                nodeModel.width = this.nodeUtils.calculateNodeWidth(nodeModel);
            }
        };
    }

    onClick(event: IClickEventArgs): void {
        const props = (event.actualObject as any)?.properties;
        if (props?.data) {
            this.toggleNodeState(props.id);
            window.parent.postMessage({
                type: 'uiEvent',
                params: {
                    type: 'click',
                    data: (event.actualObject as any).properties.data
                }
            }, '*');
        }
    }

    onDrop(event: any): void {
        if (event.source?.properties?.data && event.state === 'Start') {
            window.parent.postMessage({
                type: 'uiEvent',
                params: {
                    type: 'dragndrop',
                    data: event.source.properties.data
                }
            }, '*');
        }
    }

    onDblClick(event: IClickEventArgs): void {
    }

    addNode(): void {
        window.parent.postMessage({
            type: 'addNode'
        }, '*');
    }

    removeNode(): void {
        window.parent.postMessage({
            type: 'removeNode'
        }, '*');
    }

    nodeTransformer(node: IXplgNode): IXplgNode {
        return {};
    }

    applyTheme(theme: string): void {
        const darkTheme = [XPLGPaletteDefaultDark];
        const defaultTheme = [XPLGPaletteDefaultLight];
        const allowDarkTheme = theme === 'dark';
        const target = document.body;
        this.toggleElementClass(target, darkTheme, allowDarkTheme);
        this.toggleElementClass(target, defaultTheme, !allowDarkTheme);
    }

    toggleNodeState(targetId: string | null, selectedNodeEl?: HTMLElement): void {
        const activeStateClass = 'active';
        const parentTarget = 'advanced-node-container';
        let target = selectedNodeEl;
        if (targetId) {
            target = (document.querySelector(`#${targetId}_html_element .${parentTarget}`) as HTMLElement);
        }
        if (target && !target.classList.contains(parentTarget)) {
            target = (target.closest(`.${parentTarget}`) as HTMLElement);
        }
        if (target) {
            if (target.classList.contains(activeStateClass)) {
                this.renderer.removeClass(this.selectedNode, activeStateClass);
            } else {
                if (this.selectedNode) {
                    this.renderer.removeClass(this.selectedNode, activeStateClass);
                }
                this.selectedNode = target;
                this.renderer.addClass(this.selectedNode, activeStateClass);
            }
        }
    }

    toggleElementClass(element: Element, classList: Array<string>, isAdd?: boolean): void {
        if (isAdd) {
            classList.forEach(el => this.renderer.addClass(element, el));
        } else {
            classList.forEach(el => this.renderer.removeClass(element, el));
        }
    }

    private prepareNodes(): void {
        if (this.diagram && this.diagram.nodes.length > 1) {
            if (this.asyncDiagram) {
                let nodesWidth = 0;
                const diagramNodes = this.diagram.nodes;
                diagramNodes.forEach((node: NodeModel, index: number, nodes: NodeModel[]) => {
                    if (index >= nodes.length / 2) {
                        diagramNodes[index].tooltip = {...nodes[index].tooltip, position: 'LeftCenter'};
                    }
                    const selector = `#${node.id}_html_element .advanced-node-container:not(.hidden-node)`;
                    const bounds: any = document.querySelector(selector)?.getBoundingClientRect();
                    if (bounds) {
                        nodesWidth += bounds?.width;
                        diagramNodes[index].width = bounds?.width;
                    }
                });
                const layoutMargin = this.diagram?.layout?.margin;
                const layoutMarginLeft = layoutMargin?.left || 0;
                const layoutMarginRight = layoutMargin?.right || 0;
                const freeSpace = Math.floor(this.diagramWidth - nodesWidth - layoutMarginLeft - layoutMarginRight);
                const spaceBetweenTwoNodes = freeSpace / ((this.diagram.nodes.length - 1) / 2);
                let hiddenNodeWidth = Math.floor(spaceBetweenTwoNodes * 0.4);
                if (hiddenNodeWidth < 10) {
                    hiddenNodeWidth = 10;
                }
                for (let i = 0; i < this.diagram.nodes.length; i++) {
                    const oddNode = (i + 1) % 2;
                    if (!oddNode) {
                        this.diagram.nodes[i].minWidth = hiddenNodeWidth;
                        this.diagram.nodes[i].width = hiddenNodeWidth;
                    }
                }
                this.diagram.reset();
                this.diagram.dataBind();
                this.diagram.doLayout();
                this.diagram.fitToPage({
                    canZoomIn: false,
                    mode: 'Width',
                    margin: {
                        top: 20,
                        left: 20,
                        right: 20
                    }
                });
            } else {
                const diagramNodes = this.diagram.nodes;
                diagramNodes.forEach((node: NodeModel, index: number, nodes: NodeModel[]) => {
                    if (index >= nodes.length / 2) {
                        diagramNodes[index].tooltip = {...nodes[index].tooltip, position: 'LeftCenter'};
                    }
                    const selector = `#${node.id}_html_element .advanced-node-container`;
                    const bounds: any = document.querySelector(selector)?.getBoundingClientRect();
                    diagramNodes[index].width = bounds?.width + 8;
                });
                this.diagram.reset();
                this.diagram.dataBind();
                this.diagram.doLayout();
                if (diagramNodes.length > 2) {
                    this.diagram.fitToPage({
                        canZoomIn: false,
                        mode: 'Width',
                        margin: {
                            top: 20,
                            left: 20,
                            right: 20
                        }
                    });
                }
            }
        }
        if (this.selectedNodeId) {
            this.selectNodeById(this.selectedNodeId);
        }
    }

    private selectNodeById(id: number): void {
        if (this.asyncDiagram) {
            this.selectedNodeId = +id + (+id - 1);
        } else {
            this.selectedNodeId = id;
        }
        const parentTarget = '.advanced-node-container';
        const selectedNode: any = document.querySelector(`#diagram_htmlLayer_div > .foreign-object:nth-child(${this.selectedNodeId}) ${parentTarget}`);
        this.toggleNodeState(null, selectedNode);
    }
}
