<%@ page import="org.json.simple.parser.ParseException" %>
<%!

class GitHubJSONObjectIterable implements Iterable<JSONObject> {
	HttpClient httpClient;
	String authorizationHeader;
	String firstPageUrl;
	JSONParser jsonParser = new JSONParser();
	final String API_URI = "https://api.github.com";

	public GitHubJSONObjectIterable(HttpClient cli, String authHeader, String pathAndParams) {
		this(cli, authHeader, pathAndParams, 100);	// NOTE the documented GitHub's maximum
	}

	public GitHubJSONObjectIterable(HttpClient cli, String authHeader, String pathAndParams, int perPage) {
		httpClient = cli;
		authorizationHeader = authHeader;

		String restOfUrl = pathAndParams;
		if (!restOfUrl.contains("?")) {
			restOfUrl += "?";
		} else {
			restOfUrl += "&";
		}
		restOfUrl += "per_page=" + perPage;

		firstPageUrl = restOfUrl;
		if (!restOfUrl.startsWith(API_URI)) {
			firstPageUrl = API_URI + firstPageUrl;
		}
	}

	@Override
	public Iterator<JSONObject> iterator() {
		return new Iterator<JSONObject>() {
			JSONArray currentPage = new JSONArray();
			String nextPageUrl = firstPageUrl;
			int currentPageIndex = 0;

			@Override
			public boolean hasNext() {
				if (currentPageIndex < currentPage.size()) {
					return true;
				}
				if (nextPageUrl == null) {
					return false;
				}
				// NOTE we assume the user has called hasNext before
				HttpResponse<String> response = getGitHubHttpResponse(nextPageUrl);
				if (response.statusCode() != 200) {  // TODO handle 5xx? 400? other 4xx? 3xx??
					// NOTE do nothing, the repo has no commits probably (409 or 404)
					return false;
				}
				try {
					Object jsonBody = jsonParser.parse(response.body());
					if (jsonBody instanceof JSONArray) {
						currentPage = (JSONArray)jsonBody;
					} else {
						// NOTE like with the example of "get commit" GH does some "smart" "in-object" paging
						currentPage = new JSONArray();
						currentPage.add(jsonBody);
					}
				} catch (ParseException e) {
					throw new RuntimeException(e);
				}
				if (currentPage.size() == 0) {
					return false;
				}
				nextPageUrl = null;
				String link = response.headers().firstValue("Link").orElse(null);
				if (link != null) {
					for (String token : link.split(", ")) {
						if (token.endsWith("rel=\"next\"")) {
							nextPageUrl = token.substring(1, token.indexOf('>'));

							// NOTE not handling service denial errors, taking precautions this way instead:
							//		(assuming we're the only this GH token's user at the moment)
							String apiReqRemain =
								response.headers().firstValue("X-Ratelimit-Remaining").orElse(null);
							// NOTE this will protect from 5 `InterruptedException`s
							if (apiReqRemain != null && Integer.parseInt(apiReqRemain) < 5) {
								String apiReqRemainResetTS =
									response.headers().firstValue("X-Ratelimit-Reset").orElse(null);
								try {
									// NOTE a bold assumption here: the clock difference between GitHub's and this server is < 10s
									Thread.sleep(Integer.parseInt(apiReqRemainResetTS)*1000L - System.currentTimeMillis() + 10000);
								} catch (InterruptedException ee) {
									throw new RuntimeException(ee);
								}
							}
						}
					}
				}
				currentPageIndex = 0;

				return true;
			}

			@Override
			public JSONObject next() {
				// NOTE assuming hasNext is always called beforehand
				return (JSONObject)currentPage.get(currentPageIndex++);
			}
		};
	}

	HttpResponse<String> getGitHubHttpResponse(String url) {
		HttpRequest.Builder builder = HttpRequest.newBuilder()
			.uri(URI.create(url))
			.header("Accept", "application/vnd.github+json")
			.header("Authorization", authorizationHeader)
			// TODO? X-GitHub-Api-Version: 2022-11-28
			.GET();
		int retries = 5;
		int retryDelayMS = 1000;
		HttpResponse<String> response = null;
		while (retries > 0 && response == null) {
			try {
				response = httpClient.send(builder.build(), HttpResponse.BodyHandlers.ofString());
			} catch (IOException e) {
				if (retries != 0) {
					throw new RuntimeException(e);
				}
				retries--;
				try {
					Thread.sleep(retryDelayMS);
				} catch (InterruptedException ee) {
					throw new RuntimeException(ee);
				}
				retryDelayMS *= 2;
			} catch (InterruptedException ee) {
				throw new RuntimeException(ee);
			}
		}

		return response;
	}
}

Instant pushTime(Object pushObj) {
	return Instant.parse((String)((JSONObject)pushObj).get("created_at"));
}

Instant commitTime(Object commitObj) {
	return Instant.parse((String)((JSONObject)((JSONObject)((JSONObject)commitObj).get("commit")).get("author")).get("date"));
}

Instant auditTime(Object auditObj) {
	return Instant.ofEpochSecond((Long)((JSONObject)auditObj).get("@timestamp"));
}

%>
